import os
import aimrun
import click
import torch
import functools
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
from mltiming import timing_iterator, timing
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP, MixedPrecision, ShardingStrategy
from torch.distributed.fsdp.wrap import transformer_auto_wrap_policy
from transformers import ViTForImageClassification, ViTConfig
from tqdm import tqdm
from torch.optim import AdamW
from torch.optim.lr_scheduler import StepLR
from detonation import DeMoReplicator, FullReplicator, NoReplicator, prepare_detonation
from transformers.models.vit.modeling_vit import ViTLayer
import torch.distributed as dist


@click.command()
@click.option('--dataset', default='cifar100', type=click.Choice(['cifar10', 'cifar100']))
@click.option('--batch-size', default=32, help='input batch size for training and validation (default: 32)')
@click.option('--epochs', default=10, help='number of epochs to train (default: 10)')
@click.option('--optim', default='deto-demo', type=click.Choice(['deto-demo', 'deto-full', 'deto-none', 'adamw']))
@click.option('--optim-class', default='SGD', type=click.Choice(['SGD', 'AdamW']))
@click.option('--compression-topk', default=2)
@click.option('--compression-chunk', default=64)
@click.option('--model', default='google-t5/t5-small', type=click.Choice(['google-t5/t5-small', 'google-t5/t5-base', 'google-t5/t5-large']))
@click.option('--replicate-every', default=1)
@click.option('--skip-every', default=None, type=int)
@click.option('--device', type=click.Choice(['cpu', 'cuda', 'mps']), default='cuda')
@click.option('--shards', default=None, type=int, help="Number of shards per replication group (default: number of GPUs per node)")
def main(dataset, batch_size, epochs, optim, optim_class, compression_topk, compression_chunk, model, replicate_every, skip_every, device, shards):
    rank, nnodes, gpus = int(os.environ['RANK']), int(os.environ['NNODES']), int(os.environ['GPUS'])
    aimrun.init(repo='.', experiment='t5', args={
        'batch_size': batch_size,
        'epochs': epochs,
        'nnodes': nnodes,
        'gpus': gpus,
        'optim': optim,
        'model': model,
        'comp_topk': compression_topk if optim=='deto-demo' else None,
        'comp_chunk': compression_chunk if optim=='deto-demo' else None,
    })
    if rank == 0:
        print(aimrun.get_runs()[0].hash)
    single = device in ('cpu', 'mps') or (device == 'cuda' and nnodes == gpus == 1)
    model_and_co = setup(dataset, batch_size, optim, optim_class, compression_topk, compression_chunk, model, replicate_every, skip_every, device, single, shards)
    train(epochs, optim, single, *model_and_co)

def train(epochs, optim, single, model, train_loader, val_loader, optimizer, scheduler, train_sampler):
    rank = int(os.environ['RANK'])
    for epoch in range(1, epochs+1):
        model.train()
        train_sampler.set_epoch(epoch)
        loss_samples = torch.zeros(2).to(model.device)
        metrics = {}
        for inputs, targets in timing_iterator(iterable=tqdm(train_loader, desc=f"Training epoch {epoch}", disable=rank>0, colour="blue", ncols=150), dict=metrics, key="timing/train/fetch"):
            if single:
                batch = batch.to(model.device)
            with timing(dict=metrics, key="timing/train/zerograd"):
                optimizer.zero_grad()
            if optim == 'adamw' or single:
                with timing(dict=metrics, key="timing/train/forward"):
                    loss = model(inputs, labels=targets).loss
                with timing(dict=metrics, key="timing/train/backward"):
                    loss.backward()
            else:
                with model.no_sync(): # Disable gradient replication for the backward pass
                    with timing(dict=metrics, key="timing/train/forward"):
                        loss = model(inputs, labels=targets).loss
                    with timing(dict=metrics, key="timing/train/backward"):
                        loss.backward()
            with timing(dict=metrics, key="timing/train/optim"):
                optimizer.step()
            loss_samples[0] += loss.item()
            loss_samples[1] += len(inputs)
            with timing(dict=metrics, key="timing/train/track"):
                metrics.update({'train/loss': loss.item()})
                aimrun.track(metrics)
        # print training statistics
        if not single:
            dist.all_reduce(loss_samples, op=dist.ReduceOp.SUM)
        if rank == 0:
            train_loss = loss_samples[0] / loss_samples[1]
            print(f"Epoch {epoch} training loss  : {train_loss:.4f}")
            aimrun.track({'epoch/train/loss': train_loss}, step=epoch)
        # validation
        model.eval()
        loss_samples.zero_()
        correct = 0
        total = 0
        with torch.no_grad():
            for inputs, targets in timing_iterator(iterable=tqdm(val_loader, desc=f"Validating after epoch {epoch}", disable=rank>0, colour="green", ncols=150), dict=metrics, key='timing/val/fetch'):
                inputs, targets = inputs.to(model.device), targets.to(model.device)
                outputs = model(inputs, labels=targets)
                loss = outputs.loss
                loss_samples[0] += loss.item()
                loss_samples[1] += len(inputs)
                with timing(dict=metrics, key='timing/val/track'):
                    metrics.update({'val/loss': loss.item()})
                    aimrun.track(metrics)
                _, predicted = outputs.logits.max(1)
                total += targets.size(0)
                correct += predicted.eq(targets).sum().item()
        # End of epoch - calculate global accuracy
        correct_tensor = torch.tensor(correct, device=model.device)
        total_tensor = torch.tensor(total, device=model.device)
        dist.all_reduce(total_tensor, op=dist.ReduceOp.SUM)
        global_train_acc = 100.0 * correct_tensor.item() / total_tensor.item()
        # print validation statistics
        if not single:
            dist.all_reduce(loss_samples, op=dist.ReduceOp.SUM)
        if rank == 0:
            val_loss = loss_samples[0] / loss_samples[1]
            print(f"Epoch {epoch} validation Loss: {val_loss:.4f}, Train Acc: {global_train_acc:.3f}%")
            aimrun.track({'epoch/val/loss': val_loss}, step=epoch)
            aimrun.track({'epoch/val/accuracy': global_train_acc}, step=epoch)
        scheduler.step()
    dist.barrier()
    dist.destroy_process_group()
    aimrun.close()

def setup(dataset, batch_size, optim, optim_class, compression_topk, compression_chunk, model, replicate_every, skip_every, device, single, shards):
    torch.manual_seed(42)
    config = ViTConfig(
        image_size=224,
        patch_size=16,
        num_channels=3,
        hidden_size=384,  # ViT-Small uses 384 dim
        num_hidden_layers=12,
        num_attention_heads=6,
        intermediate_size=384 * 4, # MLP size is typically 4x hidden size
        hidden_dropout_prob=0.1,
        attention_probs_dropout_prob=0.1,
        num_labels= 100 if dataset == 'cifar100' else 10
    )
    model = ViTForImageClassification(config)
    model = model.to(device)

    # Define Transforms
    transform_train = transforms.Compose([
        transforms.RandomCrop(32, padding=4),
        transforms.RandomHorizontalFlip(),
        transforms.Resize(224),  # Resize to ViT's expected input size
        transforms.ToTensor(),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2470, 0.2435, 0.2616)) if dataset == 'cifar10' else
        transforms.Normalize((0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761))
    ])
    
    transform_test = transforms.Compose([
        transforms.Resize(224),  # Resize to ViT's expected input size
        transforms.ToTensor(),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2470, 0.2435, 0.2616)) if dataset == 'cifar10' else
        transforms.Normalize((0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761))
    ])

    # Load dataset
    if dataset == 'cifar10':
        train_dataset = torchvision.datasets.CIFAR10(root='./data', train=True, download=True, transform=transform_train)
        val_dataset = torchvision.datasets.CIFAR10(root='./data', train=False, download=True, transform=transform_test)
    else:  
        train_dataset = torchvision.datasets.CIFAR100(root='./data', train=True, download=True, transform=transform_train)
        val_dataset = torchvision.datasets.CIFAR100(root='./data', train=False, download=True, transform=transform_test)
    
    train_sampler = DistributedSampler(train_dataset, shuffle=True)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=False, num_workers=4, sampler=train_sampler)
    val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False, num_workers=4, sampler=DistributedSampler(val_dataset))
    # prepare distributed training
    if device == 'cuda':
        torch.cuda.set_device(int(os.environ['LOCAL_RANK']))
    
    auto_wrap_policy = functools.partial(transformer_auto_wrap_policy, transformer_layer_cls={ViTLayer})
    mixed_precision = MixedPrecision(param_dtype=torch.bfloat16, reduce_dtype=torch.bfloat16, buffer_dtype=torch.bfloat16) if torch.cuda.is_bf16_supported() else None
    if single:
        model = model.to(device)
        optimizer = AdamW(model.parameters(), lr=1e-3, weight_decay=0.)
    elif optim.startswith('deto-'):
        if optim == 'deto-demo':
            replicator = DeMoReplicator(compression_topk=compression_topk, compression_chunk=compression_chunk)
        elif optim == 'deto-full':
            replicator = FullReplicator()
        else:
            replicator = NoReplicator()
        model, optimizer = prepare_detonation(model, replicator, fsdp_kwargs={"auto_wrap_policy": auto_wrap_policy, "mixed_precision": mixed_precision}, replicate_every=replicate_every, skip_every=skip_every, sharding_group_size=shards)
    optim = optimizer._optimizer if hasattr(optimizer, "_optimizer") else optimizer
    scheduler = StepLR(optim, step_size=1, gamma=0.85)
    return model, train_loader, val_loader, optimizer, scheduler, train_sampler

if __name__ == '__main__':
    main()


import torch
import torch.distributed as dist
import torch.fft

from typing import Dict, Any

from .replicator import Replicator
from ..util import DCTCompress, DCTTransform

__all__ = ["DeMoReplicator"]

class DeMoReplicator(Replicator):

    def __init__(
        self,
        compression_decay: float = 0.999,
        compression_topk: int = 32,
        compression_chunk: int = 64,
    ):
        self.compression_decay = compression_decay
        self.compression_chunk = compression_chunk
        self.compression_topk = compression_topk

        if self.compression_topk <= 0:
            raise ValueError("topk_size has to be positive")
        if self.compression_chunk <= 0:
            raise ValueError("chunk_size has to be positive")
        if self.compression_decay < 0:
            raise ValueError("Negative compression_decay is currently not supported")
        if self.compression_decay >= 1:
            raise ValueError("Values of compression_decay bigger or equal to 1.0 is currently not supported")

    def init(
            self,
            optim: torch.optim.Optimizer,
            replication_parallel_group: dist.ProcessGroup | None = None,
        ):
        for group in optim.param_groups:
            for p in group["params"]:
                if p.requires_grad:
                    optim.state[p]["demo_delta"] = torch.zeros_like(p)
        self.transform = DCTTransform(optim.param_groups, self.compression_chunk)
        self.replication_parallel_group = optim.replication_parallel_group if replication_parallel_group is None else replication_parallel_group
        self._replication_world_size = self.replication_parallel_group.size()
        self.data_transmitted = []
        self.data_received = []

    def pre_step(self):
        self.data_transmit = 0
        self.data_receive = 0

    def post_step(self):
        self.data_transmitted.append(self.data_transmit)
        self.data_received.append(self.data_receive)

    def replicate(
        self,
        sharded_grad: torch.Tensor,
        param: torch.nn.Parameter,
        param_state_dict: dict,
        param_group: Dict[str, Any],
    ) -> torch.Tensor:
        # Decay delta and add the gradient
        delta = param_state_dict["demo_delta"]
        if self.compression_decay != 1:
            delta.mul_(self.compression_decay)
        delta.add_(sharded_grad, alpha=param_group["lr"])

        # Replicating delta only if needed
        if self._replication_world_size == 1:
            new_grad = delta.clone()
            delta.zero_()
            return new_grad
    
        # Compress delta
        sparse_idx, sparse_val, xshape = DCTCompress.compress(
            self.transform.encode(delta), self.compression_topk
        )

        # Estimate transmitted delta
        transmit_grad = self.transform.decode(
            DCTCompress.decompress(sparse_idx, sparse_val, xshape, param.device, param.dtype)
        )

        # Remove transmitted from delta
        delta.sub_(transmit_grad)

        # Prepare and gather the indices and values
        sparse_idx_gather = [torch.zeros_like(sparse_idx) for _ in range(self._replication_world_size)]
        sparse_val_gather = [torch.zeros_like(sparse_val) for _ in range(self._replication_world_size)]
        sparse_idx_handle = dist.all_gather(sparse_idx_gather, sparse_idx, group=self.replication_parallel_group, async_op=True)
        sparse_val_handle = dist.all_gather(sparse_val_gather, sparse_val, group=self.replication_parallel_group, async_op=True)
        sparse_idx_handle.wait()
        sparse_val_handle.wait()

        # Log I/O data size
        self.data_transmit += sparse_idx.nbytes + sparse_val.nbytes
        for si, v in zip(sparse_idx_gather, sparse_val_gather):
            self.data_receive += si.nbytes + v.nbytes

        # Decode new gradient from all nodes
        new_grad = self.transform.decode(
            DCTCompress.batch_decompress(sparse_idx_gather, sparse_val_gather, xshape, param.device, param.dtype)
        )
        return new_grad



import os
import torch
from torch.distributed.device_mesh import init_device_mesh
from torch.distributed.fsdp import (
    FullyShardedDataParallel as FSDP,
    ShardingStrategy
)
from typing import Optional, Tuple

from .optim import DeToSGD, DeToAdamW
from .repl import Replicator, DeMoReplicator

__all__ = ["prepare_detonation"]

def prepare_detonation(
    model: torch.nn.Module,
    replicator: Optional[Replicator] = DeMoReplicator(),
    sharding_group_size: Optional[int] = None,
    replication_group_size: Optional[int] = None,
    sharding_parallel_group: Optional[torch.distributed.ProcessGroup] = None,
    replication_parallel_group: Optional[torch.distributed.ProcessGroup] = None,
    fsdp_kwargs: dict = {},
    **detonation_kwargs,
) -> Tuple[torch.nn.Module, torch.optim.Optimizer]:
    world_size = int(os.environ['WORLD_SIZE'])
    local_world_size = int(os.environ['LOCAL_WORLD_SIZE'])
    if (sharding_parallel_group is None) ^ (replication_parallel_group is None):
        raise ValueError("Cannot specify only one of replication_parallel_group and sharding_parallel_group")
    if sharding_parallel_group is not None:
        if replication_group_size is not None or sharding_group_size is not None:
            raise ValueError("Cannot specify both group sizes and parallel groups")
        sharding_group_size = sharding_parallel_group.size()
        replication_group_size = replication_parallel_group.size()
    else:
        if sharding_group_size is None:
            if replication_group_size is None:
                sharding_group_size = local_world_size
                replication_group_size = world_size // local_world_size
            else:
                sharding_group_size = world_size // replication_group_size
        elif replication_group_size is None:
            replication_group_size = world_size // sharding_group_size
        mesh_2d = init_device_mesh(
            device_type="cuda",
            mesh_shape=(replication_group_size, sharding_group_size),
        )
        sharding_parallel_group = mesh_2d.get_group(1)
        replication_parallel_group = mesh_2d.get_group(0)
    assert world_size == sharding_group_size * replication_group_size
    assert local_world_size % sharding_group_size == 0
    model = FSDP(
        model,
        process_group=(sharding_parallel_group, replication_parallel_group),
        device_id=int(os.environ['LOCAL_RANK']),
        sharding_strategy=ShardingStrategy.HYBRID_SHARD,
        **fsdp_kwargs,
    )
    optim = DeToSGD(
        model.parameters(),
        replicator=replicator,
        sharding_parallel_group=sharding_parallel_group,
        replication_parallel_group=replication_parallel_group,
        **detonation_kwargs,
    )
    return model, optim



import aimrun
import click
from datasets import load_dataset
from detonation import DeMoReplicator, FullReplicator, NoReplicator, prepare_detonation
import functools
from mltiming import timing_iterator, timing
import os
import torch
import torch.distributed as dist
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP, MixedPrecision, ShardingStrategy
from torch.distributed.fsdp.wrap import transformer_auto_wrap_policy
from torch.optim import AdamW
from torch.optim.lr_scheduler import StepLR
from torch.utils.data import DataLoader, Dataset
from torch.utils.data.distributed import DistributedSampler
from tqdm import tqdm
from transformers import T5Config, T5ForConditionalGeneration, T5Tokenizer
from transformers.models.t5.modeling_t5 import T5Block


@click.command()
@click.option('--batch-size', default=32, help='input batch size for training and validation (default: 32)')
@click.option('--epochs', default=10, help='number of epochs to train (default: 10)')
@click.option('--optim', default='deto-demo', type=click.Choice(['deto-demo', 'deto-full', 'deto-none', 'adamw']))
@click.option('--optim-class', default='SGD', type=click.Choice(['SGD', 'AdamW']))
@click.option('--compression-topk', default=2)
@click.option('--compression-chunk', default=64)
@click.option('--model', default='google-t5/t5-small', type=click.Choice(['google-t5/t5-small', 'google-t5/t5-base', 'google-t5/t5-large']))
@click.option('--replicate-every', default=1)
@click.option('--skip-every', default=None, type=int)
@click.option('--device', type=click.Choice(['cpu', 'cuda', 'mps']), default='cuda')
@click.option('--shards', default=None, type=int, help="Number of shards per replication group (default: number of GPUs per node)")
def main(batch_size, epochs, optim, optim_class, compression_topk, compression_chunk, model, replicate_every, skip_every, device, shards):
    rank, nnodes, gpus = int(os.environ['RANK']), int(os.environ['NNODES']), int(os.environ['GPUS'])
    aimrun.init(repo='.', experiment='t5', args={
        'batch_size': batch_size,
        'epochs': epochs,
        'nnodes': nnodes,
        'gpus': gpus,
        'optim': optim,
        'model': model,
        'comp_topk': compression_topk if optim=='deto-demo' else None,
        'comp_chunk': compression_chunk if optim=='deto-demo' else None,
    })
    if rank == 0:
        print(aimrun.get_runs()[0].hash)
    single = device in ('cpu', 'mps') or (device == 'cuda' and nnodes == gpus == 1)
    model_and_co = setup(batch_size, optim, optim_class, compression_topk, compression_chunk, model, replicate_every, skip_every, device, single, shards)
    train(epochs, optim, single, *model_and_co)

def train(epochs, optim, single, model, train_loader, val_loader, optimizer, scheduler, train_sampler):
    rank = int(os.environ['RANK'])
    for epoch in range(1, epochs+1):
        # train
        model.train()
        #prof.export_chrome_trace('trace.json')
        train_sampler.set_epoch(epoch)
        loss_samples = torch.zeros(2).to(model.device)
        metrics = {}
        for batch in timing_iterator(iterable=tqdm(train_loader, desc=f"Training epoch {epoch}", disable=rank>0, colour="blue", ncols=150), dict=metrics, key="timing/train/fetch"):
            if single:
                batch["source_ids"] = batch["source_ids"].to(model.device)
                batch["source_mask"] = batch["source_mask"].to(model.device)
                batch["target_ids"] = batch["target_ids"].to(model.device)
            with timing(dict=metrics, key="timing/train/zerograd"):
                optimizer.zero_grad()
            if optim == 'adamw' or single:
                with timing(dict=metrics, key="timing/train/forward"):
                    loss = model(input_ids=batch["source_ids"],attention_mask=batch["source_mask"],labels=batch["target_ids"] )["loss"]
                with timing(dict=metrics, key="timing/train/backward"):
                    loss.backward()
            else:
                with model.no_sync(): # Disable gradient replication for the backward pass
                    with timing(dict=metrics, key="timing/train/forward"):
                        loss = model(input_ids=batch["source_ids"],attention_mask=batch["source_mask"],labels=batch["target_ids"] )["loss"]
                    with timing(dict=metrics, key="timing/train/backward"):
                        loss.backward()
            with timing(dict=metrics, key="timing/train/optim"):
                optimizer.step()
            loss_samples[0] += loss.item()
            loss_samples[1] += len(batch)
            with timing(dict=metrics, key="timing/train/track"):
                metrics.update({'train/loss': loss.item()})
                aimrun.track(metrics)
                metrics.clear()
        # print training statistics
        if not single:
            dist.all_reduce(loss_samples, op=dist.ReduceOp.SUM)
        if rank == 0:
            train_loss = loss_samples[0] / loss_samples[1]
            print(f"Epoch {epoch} training loss  : {train_loss:.4f}")
            aimrun.track({'epoch/train/loss': train_loss}, step=epoch)
        # validate
        model.eval()
        loss_samples.zero_()
        metrics.clear()
        with torch.no_grad():
            for batch in timing_iterator(iterable=tqdm(val_loader, desc=f"Validating after epoch {epoch}", disable=rank>0, colour="green", ncols=150), dict=metrics, key='timing/val/fetch'):
                if single:
                    batch["source_ids"] = batch["source_ids"].to(model.device)
                    batch["source_mask"] = batch["source_mask"].to(model.device)
                    batch["target_ids"] = batch["target_ids"].to(model.device)
                with timing(dict=metrics, key='timing/val/forward'):
                    loss = model(input_ids=batch["source_ids"],attention_mask=batch["source_mask"],labels=batch["target_ids"])["loss"]
                loss_samples[0] += loss.item()
                loss_samples[1] += len(batch)
                with timing(dict=metrics, key='timing/val/track'):
                    metrics.update({'val/loss': loss.item()})
                    aimrun.track(metrics)
                    metrics.clear()
        # print validation statistics
        if not single:
            dist.all_reduce(loss_samples, op=dist.ReduceOp.SUM)
        if rank == 0:
            val_loss = loss_samples[0] / loss_samples[1]
            print(f"Epoch {epoch} validation Loss: {val_loss:.4f}")
            aimrun.track({'epoch/val/loss': val_loss}, step=epoch)
        scheduler.step()
    dist.barrier()
    dist.destroy_process_group()
    aimrun.close()

def setup(batch_size, optim, optim_class, compression_topk, compression_chunk, model, replicate_every, skip_every, device, single, shards):
    torch.manual_seed(42)
    # prepare model
    tokenizer =  T5Tokenizer.from_pretrained(model, legacy=False)
    model = T5ForConditionalGeneration(T5Config.from_pretrained(model))
    # prepare dataset
    train_test_split = load_dataset("gursi26/wikihow-cleaned", split="train").train_test_split(test_size=0.2)
    train_dataset = WikiHow(tokenizer, train_test_split['train'], num_samples=15000)
    val_dataset = WikiHow(tokenizer, train_test_split['test'], num_samples=3000)
    train_sampler = DistributedSampler(train_dataset, shuffle=True)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, sampler=train_sampler)
    val_loader = DataLoader(val_dataset, batch_size=batch_size, sampler=DistributedSampler(val_dataset))
    # prepare distributed training
    if device == 'cuda':
        torch.cuda.set_device(int(os.environ['LOCAL_RANK']))
    auto_wrap_policy = functools.partial(transformer_auto_wrap_policy, transformer_layer_cls={T5Block})
    mixed_precision = MixedPrecision(param_dtype=torch.bfloat16, reduce_dtype=torch.bfloat16, buffer_dtype=torch.bfloat16) if torch.cuda.is_bf16_supported() else None
    if single:
        model = model.to(device)
        optimizer = AdamW(model.parameters(), lr=1e-3, weight_decay=0.)
    elif optim.startswith('deto-'):
        if optim == 'deto-demo':
            replicator = DeMoReplicator(compression_topk=compression_topk, compression_chunk=compression_chunk)
        elif optim == 'deto-full':
            replicator = FullReplicator()
        else:
            replicator = NoReplicator()
        model, optimizer = prepare_detonation(model, replicator, fsdp_kwargs={"auto_wrap_policy": auto_wrap_policy, "mixed_precision": mixed_precision}, replicate_every=replicate_every, skip_every=skip_every, sharding_group_size=shards, optim_class=optim_class)
    else:
        model = FSDP(model, auto_wrap_policy=auto_wrap_policy, mixed_precision=mixed_precision, device_id=int(os.environ['LOCAL_RANK']), sharding_strategy=ShardingStrategy.HYBRID_SHARD)
        optimizer = AdamW(model.parameters(), lr=1e-3, weight_decay=0.)
    optim = optimizer._optimizer if hasattr(optimizer, "_optimizer") else optimizer
    scheduler = StepLR(optim, step_size=1, gamma=0.85)
    return model, train_loader, val_loader, optimizer, scheduler, train_sampler

class WikiHow(Dataset):
    def __init__(self, tokenizer, dataset, num_samples):
        self.dataset, self.tokenizer = dataset.select(list(range(0, num_samples))), tokenizer
    def __len__(self):
        return self.dataset.shape[0]
    def __getitem__(self, index):
        source = self.dataset[index]['text']
        target = self.dataset[index]['title']
        if not source or not target:
            source = "because of this , it is recommended to increase your intake of vitamin c while you are dealing with a virus . aside from taking a vitamin c supplement , you can also eat fruits that have high amounts of vitamin c . these include grapefruit , kiwi , strawberries , lemon , lime , blackberries , oranges , papaya , pineapple , pomelo , and raspberries . eat vegetables that are rich in vitamin c . these include brussel sprouts , broccoli , onions , garlic , red and green peppers , tomatoes , and radishes . you can also consider making vegetable soup , if you dont like eating raw veggies . if you have ever wondered why people always give their kids chicken noodle soup when they are sick , its because chicken soup is a wonder when it comes to recovering from a virus . not only does chicken soup act as an anti - inflammatory , it also temporarily helps to relieve congestion by unblocking your nasal passages . you can also add onions , garlic , and other veggies to your soup to boost its vitamin and mineral count . zinc governs enzymes in our body that activate different parts of our immune system that fight against infection . most people choose to take a 25 mg zinc supplement before one meal each day , but you can also add zinc - rich foods to your diet . these foods include spinach , mushrooms , beef , lamb , pork or chicken , and cooked oysters . zinc has been shown to be most effective when taken for two to three days at the beginning of a cold or flu . start taking zinc as soon as you think you may be getting sick . you can also purchase lozenges that contain zinc , which you can suck on . you can buy these and other zinc supplements at your local pharmacy . do not take zinc supplements if you take antibiotics such as tetracyclines , fluoroquinolones , penicillamine a drug used in wilsons disease , or cisplatin a medication used in cancer , due to the fact that zinc decreases the efficiency of these drugs . echinacea is a type of plant that is often made into a tea or taken as a supplement . when consumed , it helps to increase the number of leukocytes white blood cells that boost your immunity and other immune - related cells in your body . you can consume echinacea by drinking tea or juice made from the plant , or by taking supplements bought at a pharmacy or health foods store . other natural remedies to consider include eucalyptus , elderberry , honey , and reishi and shiitake mushrooms ."
            target = "how to treat a viral infection 2"
        for sub in ('Example of text:', 'Example of Summary:', '\n', '``', '"'):
            source, target = source.replace(sub, ''), target.replace(sub, '')
        source = self.tokenizer.batch_encode_plus([source], max_length=512, padding='max_length', truncation=True, return_tensors="pt")
        targets = self.tokenizer.batch_encode_plus([target], max_length=150, padding='max_length', truncation=True, return_tensors="pt")
        return {"source_ids": source['input_ids'].squeeze(), "source_mask": source['attention_mask'].squeeze(), "target_ids": targets['input_ids'].squeeze(), "target_mask": targets['attention_mask'].squeeze()}

if __name__ == '__main__':
    main()



import torch
import torch.distributed as dist
from torch.optim import SGD
from typing import Callable, List

from ..detonatiomixin import DeToNATIONMixin
from ..repl import DeMoReplicator, Replicator

__all__ = ["DeToSGD"]

class DeToSGD(SGD, DeToNATIONMixin):

    def __init__(
        self,
        params,
        detonation_weight_decay: float = 0.0,
        detonation_sign: bool = True,
        sharding_parallel_group: dist.ProcessGroup | None = None,
        replication_parallel_group: dist.ProcessGroup | List[dist.ProcessGroup] | None = None,
        replicator: Replicator | List[Replicator] = DeMoReplicator(),
        replicate_every: int | List[int] = 1,
        skip_every: int | List[int] | None = None,
        *args,
        **kwargs,
    ):
        SGD.__init__(
            self,
            params,
            *args,
            **kwargs,
        )
        DeToNATIONMixin.__init__(
            self,
            detonation_weight_decay=detonation_weight_decay,
            detonation_sign=detonation_sign,
            sharding_parallel_group=sharding_parallel_group,
            replication_parallel_group=replication_parallel_group,
            replicator=replicator,
            replicate_every=replicate_every,
            skip_every=skip_every,
        )

    @torch.no_grad()
    def step(self, closure: Callable | None = None):
        return DeToNATIONMixin.step(self, closure=closure, base_step=SGD.step)



import torch
import torch.distributed as dist
from typing import Any, Dict

from .replicator import Replicator

__all__ = ["FullReplicator"]

class FullReplicator(Replicator):

    def init(
            self,
            optim: torch.optim.Optimizer,
            replication_parallel_group: dist.ProcessGroup | None = None,
        ):
        self.replication_parallel_group = optim.replication_parallel_group if replication_parallel_group is None else replication_parallel_group
        self.data_transmitted = []
        self.data_received = []

    def pre_step(self):
        self.data_transmit = 0
        self.data_receive = 0

    def post_step(self):
        self.data_transmitted.append(self.data_transmit)
        self.data_received.append(self.data_receive)

    def replicate(
        self,
        sharded_grad: torch.Tensor,
        param: torch.nn.Parameter,
        param_state_dict: dict,
        param_group: Dict[str, Any],
        step_metrics: dict,
    ) -> torch.Tensor:
        dist.all_reduce(sharded_grad, dist.ReduceOp.AVG, group=self.replication_parallel_group)
        self.data_receive += sharded_grad.nbytes
        self.data_transmit += sharded_grad.nbytes
        return sharded_grad.to(param.device).to(param.dtype)
